from .project import project_on_rhythm, project_on_one_chord,\
    get_absolute_voices, reproject_on_multiple_chords, get_absolute_voice
import numpy as np


def create_counterpoint(fixed_voices, voices):
    """
    Create a counterpoint given a list of fixed voices and voices to adapt.
    You should provide all the melodies, the counterpoint only mutate existing voices to fit best the fixed voices.
    It tries to maximize a score, greedily, note per note and voice per voice. This score uses the following rules :

    - No consecutives fifths or octaves or unissons
    - No hidden fifths or octaves
    - No parallel dissonnances
    - No triton (s3 and s6)
    - Try to avoid moving a note too far appart


    Parameters
    ----------
    fixed_voices : list[Melody]
                   Voices that serves as cantus firmus
        
    voices :  list[Melody]
              Voices that will be modified to fit the cantus firmus.
              Only the notes will be changed, never the rythm.
        

    Returns
    -------
    result: list[Melody] shape of voices
            The counterpointed voices

    """

    # Convert fixed_voices in absolute
    fixed_voices = get_absolute_voices(fixed_voices)

    # The list of subjects that will grow with each corrected voice
    melody_subjects = fixed_voices

    result = []
    for voice in voices:
        # Remove absolute, Project all on rythm
        voice = get_absolute_voice(voice)
        subjects = get_projections_on_voice(melody_subjects, voice)
        voice_to_fix = get_array(voice)
        fixed_voice = get_counterpoint(subjects, voice_to_fix)
        melody_fixed = convert_array_to_melody(voice, fixed_voice)
        melody_subjects.append(melody_fixed)
        result.append(melody_fixed)
        # Convert back to melody

    return result


def create_counterpoint_on_score(score, fixed_parts, counterpoint_parts=None):
    """
    Create a counterpoint on parts for a score
    - Put everything on the same chord
    - Call :func:`~create_counterpoint` on melodies
    - Put everything back in the original chord progression

    You should provide all the melodies, the counterpoint only mutate existing voices to fit best the fixed voices.
    It tries to maximize a score, greedily, note per note and voice per voice. This score uses the following rules :

    - No consecutives fifths or octaves or unissons
    - No hidden fifths or octaves
    - No parallel dissonnances
    - No triton (s3 and s6)
    - Try to avoid moving a note too far appart

    Parameters
    ----------
    score: Score
    fixed_parts: list[str]
    Name of parts that are fixed  (eg: piano__0)
    counterpoint_parts: list[str] or None
        Name of parts that will move (eg: piano__1)
        If None, all the remaining voices will be applied a counterpoint on the fixed_parts

    Returns
    -------
    score: Score
           Resulting score

    """
    from musiclang import Score
    if counterpoint_parts is None:
        counterpoint_parts = set(score.instruments) - set(fixed_parts)

    # Deal with bass and chord notes
    chord, _, chords_offsets, _, chords = project_on_one_chord(score)
    fixed_voices = [chord.score[part] for part in fixed_parts]
    moving_voices = [chord.score[part] for part in counterpoint_parts]
    moving_voices = create_counterpoint(fixed_voices, moving_voices)
    for voice, part in zip(moving_voices, counterpoint_parts):
        chord.score[part] = voice
    projection = chord.to_score().project_on_score(score)
    return projection


def create_counterpoint_on_chord(chord, subject_parts, counterpoint_parts):
    """

    Parameters
    ----------
    chord :
        
    subject_parts :
        
    counterpoint_parts :
        

    Returns
    -------

    """
    fixed_voices = [chord.score[chord.parts[idx]] for idx in subject_parts]
    voices = [chord.score[chord.parts[idx]] for idx in counterpoint_parts]
    new_voices = create_counterpoint(fixed_voices, voices)
    return chord(**{**chord.score, **{chord.parts[idx]: voice for idx, voice in zip(counterpoint_parts, new_voices)}})



def get_array_fixed(voice):
    """

    Parameters
    ----------
    voice :
        

    Returns
    -------

    """
    prev = None
    res = []
    for n in voice.notes:
        if n.is_note:
            prev = n.val + 7 * n.octave
            res.append(prev)
        elif n.is_silence:
            res.append(None)
        elif n.is_continuation:
            res.append(prev)
    return res

def get_array(voice):
    """

    Parameters
    ----------
    voice :
        

    Returns
    -------

    """
    return [n.val + 7 * n.octave if n.is_note else None for n in voice.notes]

def get_projections_on_voice(voices, reference):
    """Project all voices on rythm given by reference
    and convert it to array of values

    Parameters
    ----------
    voices :
        param reference:
    reference :
        

    Returns
    -------

    """
    return [get_array_fixed(project_on_rhythm(reference, voice)) for voice in voices]


AUTHORIZED_INTERVALS = [0, 2, 3, 4, 5]
FORBIDDEN_PARALLELS = [0, 4, 3, 7]


def get_future_delta(delta):
    """Explore around (0, 1, -1, 2, -2, etc ...)

    Parameters
    ----------
    delta :
        return:

    Returns
    -------

    """
    if delta == 0:
        return 1
    elif delta > 0:
        return -delta
    else:
        return -delta + 1

def get_delta_list():
    """ """
    deltas = [0, 1, -1, 2, -2, 3, -3, 4, -4]
    #deltas = [0, 1, -2, 2, -2]
    np.random.shuffle(deltas)
    return deltas

def is_dissonnance(n1, n2):
    """

    Parameters
    ----------
    n1 :
        
    n2 :
        

    Returns
    -------

    """
    return interval(n1, n2) not in AUTHORIZED_INTERVALS

def is_forbidden_parallel(last_interval, subject, candidate):
    """

    Parameters
    ----------
    last_interval :
        
    subject :
        
    candidate :
        

    Returns
    -------

    """
    new_interval = interval(subject, candidate)
    return new_interval in FORBIDDEN_PARALLELS and last_interval == new_interval

def is_parallel_dissonnance(last_interval, subject, candidate):
    """

    Parameters
    ----------
    last_interval :
        
    subject :
        
    candidate :
        

    Returns
    -------

    """
    return is_dissonnance(subject, candidate) and last_interval not in AUTHORIZED_INTERVALS


def nb_forbidden_parallel(last_intervals, subject_notes, candidate):
    """

    Parameters
    ----------
    last_intervals :
        
    subject_notes :
        
    candidate :
        

    Returns
    -------

    """
    count = 0
    for s, inter in zip(subject_notes, last_intervals):
        if s is None:
            continue
        elif inter is None:
            continue
        elif is_forbidden_parallel(inter, s, candidate):
            count += 1
    return count

def nb_parallel_dissonnances(last_intervals, subject_notes, candidate):
    """

    Parameters
    ----------
    last_intervals :
        
    subject_notes :
        
    candidate :
        

    Returns
    -------

    """
    count = 0
    for s, inter in zip(subject_notes, last_intervals):
        if s is None:
            continue
        elif inter is None:
            continue
        elif is_parallel_dissonnance(inter, s, candidate):
            count += 1
    return count



def is_triton(candidate, subject_note):
    """

    Parameters
    ----------
    candidate :
        
    subject_note :
        

    Returns
    -------

    """
    return {candidate % 7, subject_note % 7} == {3, 6}

def scorer(subject_notes, note, delta, last_intervals, last_notes):
    """

    Parameters
    ----------
    subject_notes :
        
    note :
        
    delta :
        
    last_intervals :
        

    Returns
    -------

    """

    not_silenced_subject_notes = [s for s in subject_notes if s is not None]
    candidate = note + delta
    NB_DISSONNANCES = sum([1 * is_dissonnance(candidate, s) for s in not_silenced_subject_notes])
    NB_FORBIDDEN_PARALLELS = nb_forbidden_parallel(last_intervals[-1], subject_notes, candidate)
    NB_PARALLEL_DISSONNANCES = nb_parallel_dissonnances(last_intervals[-1], subject_notes, candidate)
    NB_TRITON = sum([1 * is_triton(candidate, s) for s in not_silenced_subject_notes])
    DISTANCE = abs(delta)
    LAST_NOTE_SAME = 1.0 * (candidate == last_notes[-1]) if len(last_notes) > 0 else 0.0
    score = 10 - 3.0 * NB_DISSONNANCES
    score += - 2.0 * NB_TRITON
    score += - 4 * NB_FORBIDDEN_PARALLELS - 4 * NB_PARALLEL_DISSONNANCES - 0.5 * DISTANCE
    score += - 2.0 * LAST_NOTE_SAME
    return score

def get_counterpoint_for_one_note(subjects_notes, note, last_intervals, last_notes, delta=0):
    """

    Parameters
    ----------
    subjects_notes :
        
    note :
        
    last_intervals :
        
    delta :
         (Default value = 0)

    Returns
    -------

    """

    deltas = get_delta_list()
    scores = [scorer(subjects_notes, note, delta, last_intervals, last_notes) for delta in deltas]
    max_score, max_score_idx = np.max(scores), np.argmax(scores)
    best_delta = deltas[max_score_idx]

    chosen_candidate = note + best_delta
    assert len(subjects_notes) == len(last_intervals[-1]), "{} {}".format(len(subjects_notes), len(last_intervals[-1]))
    last_intervals.append([interval(s, chosen_candidate, replace=old_inter) for s, old_inter in zip(subjects_notes, last_intervals[-1])])
    return chosen_candidate, last_intervals, max_score



def interval(n1, n2, replace=None):
    """

    Parameters
    ----------
    n1 :
        
    n2 :
        
    replace :
         (Default value = None)

    Returns
    -------

    """
    if n1 is None or n2 is None:
        if replace:
            return replace
        return None
    return abs(n2 - n1) % 7


def clear_list_intervals(intervals):
    """

    Parameters
    ----------
    intervals :
        

    Returns
    -------

    """
    if len(intervals) > 0:
        return intervals[-3:]
    return intervals


def get_counterpoint(subjects, to_fix):
    """Given a list of subjects modify the voice to best fit a counterpoint

    Parameters
    ----------
    fixed_voices :
        param voice:
    subjects :
        
    to_fix :
        

    Returns
    -------

    """
    subjects = np.asarray(subjects)
    last_intervals = [[None for s in subjects]]
    result = []
    total_score = 0
    last_notes = []
    for i, n in enumerate(to_fix):
        if n is None:
            result.append(None)
        else:
            new_note, last_intervals, max_score = get_counterpoint_for_one_note(subjects[:, i].tolist(), n, last_intervals, last_notes)
            result.append(new_note)
            total_score += max_score
            last_notes.append(new_note)

        last_intervals = clear_list_intervals(last_intervals)

    return result


def convert_array_to_melody(rythm, notes):
    """

    Parameters
    ----------
    rythm :
        
    notes :
        

    Returns
    -------

    """
    new_melody = None
    for idx, note in enumerate(rythm.notes):
        to_add = note.copy()
        if note.is_note:
            to_add.val = 0
            to_add.type = "s"
            to_add.octave = 0
            to_add.octave = notes[idx] // 7
            to_add.val = notes[idx] % 7

        new_melody += to_add

    return new_melody




from .pipeline import TransformPipeline, ConcatPipeline
from .mask import Mask
from .transformer import Transformer, ChordTransformer, ScoreTransformer, MelodyTransformer, NoteTransformer
from .transformer import ChordFilter, ChordFilterTransform, MelodyFilterTransform, MelodyFilter
from .transformer import NoteFilter, NoteFilterTransform, FeatureExtractor
from .transformer import MaskFilter, NoteMaskFilter, ChordMaskFilter, MelodyMaskFilter
from .merger import ScoreMerger
from .graph import TransformGraph
from .composing import VoiceLeading, PartComposer, create_counterpoint_on_score
from .dynamics import PitchDynamizer
from .melody import ContinuationWhenSameNote
from .composing import Patternator, OrchestralLayer, MelodicLayer, GlobalLayer
from .orchestrations import get_epic_orchestration, get_nocturne_orchestration

__all__ = ['TransformPipeline', 'ConcatPipeline', 'Mask',
           'Transformer', 'ChordTransformer', 'ScoreTransformer', 'MelodyTransformer',
           'NoteTransformer',
           'ChordFilter', 'ChordFilterTransform', 'MelodyFilterTransform', 'MelodyFilter',
           'NoteFilter', 'NoteFilterTransform', 'FeatureExtractor', 'MaskFilter', 'NoteMaskFilter',
           'ChordMaskFilter', 'MelodyMaskFilter', 'ScoreMerger', 'TransformGraph',
           'VoiceLeading', 'PitchDynamizer', 'ContinuationWhenSameNote', 'Patternator',
           'get_epic_orchestration', 'get_nocturne_orchestration', 'PartComposer','OrchestralLayer',
              'MelodicLayer', 'GlobalLayer', 'create_counterpoint_on_score'
           ]


import numpy as np

from .base_transformer import Transformer
from .mask import Mask


class NoteTransformer(Transformer):
    TYPE = 'NOTE'

    def action(self, note: 'Note', chord=None, instrument=None, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        from musiclang import Score, Melody, Chord, Note
        if isinstance(element, Note):
            return self.action(element, **kwargs) if on(element, **kwargs) else element.copy()
        if isinstance(element, Melody):
            return self.apply_on_melody(element, on=on, **kwargs)
        elif isinstance(element, list):
            return [self(n) for n in element]
        elif isinstance(element, Chord):
            return self.apply_on_chord(element, on=on, **kwargs)
        elif isinstance(element, Score):
            return self.apply_on_score(element, on=on, **kwargs)
        else:
            raise Exception(f'Cannot apply to type {element.__class__}')


class NoteFilterTransform(NoteTransformer):

    def get_default(self, element):
        return None


class MaskFilter(NoteFilterTransform):

    def __init__(self, on):
        self.on = on

    def action(self, note, chord=None, instrument=None, **kwargs):
        return note

    def __call__(self, element, on=Mask(), **kwargs):
        return super().__call__(element, on=self.on & on, **kwargs)

class NoteMaskFilter(MaskFilter):
    pass


class NoteFilter(NoteFilterTransform):

    def action(self, note, chord=None, instrument=None, **kwargs):
        return note

    def filter(self, element, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        query_filter = Mask.Note() > Mask.Func(lambda x, **k: self.filter(element, **k))
        return super().__call__(element, on=query_filter & on, **kwargs)


class MelodyTransformer(Transformer):
    TYPE = 'MELODY'

    def action(self, melody, chord=None, instrument=None, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        from musiclang import Score, Melody, Chord, Note
        if isinstance(element, Melody):
            return self.action(element, **kwargs)
        elif isinstance(element, Note):
            return self.action(Melody([element]), **kwargs)
        elif isinstance(element, list):
            return [self(n, **kwargs) for n in element]
        elif isinstance(element, Chord):
            return self.apply_on_chord(element, on=on, **kwargs)
        elif isinstance(element, Score):
            return self.apply_on_score(element, on=on, **kwargs)
        else:
            raise Exception(f'Cannot apply to type {element.__class__}')


class MelodyFilterTransform(MelodyTransformer):

    def get_default(self, element):
        return None


class MelodyFilter(MelodyFilterTransform):

    def action(self, melody, chord=None, instrument=None, **kwargs):
        return melody

    def filter(self, element, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        query_filter = Mask.Melody() > Mask.Func(lambda x, **k: self.filter(element, **k))
        return super().__call__(element, on=query_filter & on, **kwargs)


class MelodyMaskFilter(MelodyFilterTransform):

    def __init__(self, on):
        self.on = on

    def action(self, note, chord=None, instrument=None, **kwargs):
        return note

    def __call__(self, element, on=Mask(), **kwargs):
        return super().__call__(element, on=self.on & on, **kwargs)


class ChordTransformer(Transformer):
    TYPE = 'CHORD'

    def action(self, chord, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        from musiclang import Score, Melody, Chord
        if isinstance(element, Chord):
            return self.action(element, **kwargs)
        elif isinstance(element, Score):
            res = self.apply_on_score(element, on=on, **kwargs)
            return res
        else:
            raise Exception(f'Cannot apply on part of instrument in the type {type(element)}')


class ChordFilterTransform(ChordTransformer):

    def get_default(self, element):
        return None


class ChordFilter(ChordFilterTransform):

    def action(self, chord, **kwargs):
        return chord

    def filter(self, element, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        query_filter = Mask.Chord() > Mask.Func(lambda x, **k: self.filter(x, **k))
        return super(ChordFilter, self).__call__(element, on=query_filter & on, **kwargs)


class ChordMaskFilter(ChordFilterTransform):

    def __init__(self, on):
        self.on = on

    def action(self, note, **kwargs):
        return note

    def __call__(self, element, on=Mask(), **kwargs):
        return super().__call__(element, on=self.on & on, **kwargs)


class DictTransformer(Transformer):
    TYPE = 'DICT'

    def action(self, melody, chord=None, **kwargs):
        raise NotImplemented

    def __call__(self, element, on=Mask(), **kwargs):
        from musiclang import Score, Melody, Chord
        if isinstance(element, dict):
            return self.action(element, **kwargs)
        if isinstance(element, Chord):
            element = element.copy()
            element.score = self.action(element.score, chord=element, **kwargs)
            return element
        elif isinstance(element, Score):
            chords = [self(m, on=on.child(element), **kwargs) if on(m) else self.get_default(m) for m in element.chords]
            chords = [chord for chord in chords if chord is not None]
            return Score(chords)
        else:
            raise Exception(f'Cannot apply on part of instrument in the type {type(element)}')


class ScoreTransformer(Transformer):
    TYPE = 'SCORE'

    def action(self, score, **kwargs):
        raise NotImplemented

    def __call__(self, data, **kwargs):
        """
        Act on a serie of chords
        :param melody:
        :param kwargs:
        :return:
        """
        from musiclang import Chord
        if isinstance(data, Chord):
            data = data.to_score()
        return self.action(data, **kwargs)


class FeatureExtractor(ScoreTransformer):
    """
    Take a score in input and can returns anything
    """
    TYPE = 'FEATURE'



class Identity(ScoreTransformer):
    def __call__(self, melody, **kwargs):
        return melody.copy()





from musiclang.transform import FeatureExtractor
from collections import Counter


class ExtractMainTonality(FeatureExtractor):
    """
    Return the tonality of a given score in terms of chord frequence of apparition
    """

    def action(self, score, **kwargs):
        d_res = {}
        for idx, chord in enumerate(score):
            tone = chord.tonality
            d_res[tone] = d_res.get(tone, 0) + chord.duration * (0.96 ** idx)

        tonality = max(tuple(d_res.items()), key=lambda x: x[1])[0]
        return tonality


class MostCommonTonalities(FeatureExtractor):
    """
    Return the most common n tonalities of a given score in terms of chord frequence of apparition
    """

    def __init__(self, n):
        self.n = n

    def action(self, score, **kwargs):
        tonalities = [chord.tonality for chord in score]
        tonalities = [t[0] for t in Counter(tonalities).most_common(self.n)[:self.n]]
        return tonalities




from musiclang.write.library import *
import numpy as np

def offset_between_chords(c1, c2):
    """How much should we transpose the melody of the second chord for the s0 of c2 to be the nearest of c1's s0

    Parameters
    ----------
    c1 :
        param c2:
    c2 :
        

    Returns
    -------

    """
    from musiclang.write.constants import DEGREE_TO_SCALE_DEGREE
    offset_degrees = c2.element - c1.element

    c1_degree = c1.tonality.degree if c1.tonality is not None else 0
    c2_degree = c2.tonality.degree if c2.tonality is not None else 0
    c1_octave = c1.full_octave if c1.tonality is not None else 0
    c2_octave = c2.full_octave if c2.tonality is not None else 0
    offset_tonalities = int(np.sign(c2_degree - c1_degree) * DEGREE_TO_SCALE_DEGREE[abs(c2_degree - c1_degree)])
    offset_octave = c1_octave - c2_octave
    offset_octave_chords = c2.octave - c1.octave

    return offset_degrees + offset_tonalities + 7 * (offset_octave + offset_octave_chords)


def project_on_several_chord(score, chords):
    """

    Parameters
    ----------
    score
    chords

    Returns
    -------

    """
    from musiclang import Score
    chord, _, _, _, _ = project_on_one_chord(score)
    _, _, chords_offsets, _, _ = project_on_one_chord(chords)

    projection = chord.to_score().project_on_score(chords, keep_score=False)
    projection = Score([s & (-i) for s, i in zip(projection.chords, chords_offsets)])
    return projection

def project_on_score_keep_notes(score1, score2):
    """
    Project the score1 on the score2 keping the notes as close as possible to the original score
    For example:

    Parameters
    ----------
    score1: Score
            Score to project
    score2: Score
            Score on which to project (the resulting score will have the same chord progression)
    Returns
    -------

    """
    from musiclang import Score
    chord, _, _, _, chords = project_on_one_chord(score1)
    _, _, chords_offsets, _, _ = project_on_one_chord(score2)
    projection = chord.to_score().project_on_score(score2, voice_leading=False)
    projection = Score([s & (-i) for s, i in zip(projection.chords, chords_offsets)])
    return projection


def project_on_one_chord(score):
    """Clever projection of a score in several chords into the first chord
    :return:

    Parameters
    ----------
    score :
        

    Returns
    -------

    """
    chords = score.chords
    first_chord = chords[0]
    all_parts = list(dict.fromkeys([part for chord in score.chords for part in chord.parts]))
    idx_stops = {part: [] for part in all_parts}
    melodies = {part: None for part in all_parts}
    offsets = []
    for chord in chords:
        offset = offset_between_chords(first_chord, chord)
        offsets.append(offset)
        for part in all_parts:
            if part in chord.score.keys():
                start = idx_stops[part][-1][-1] if len(idx_stops[part]) > 0 else 0
                stop = start + chord.score[part].duration
                idx_stops[part].append((start, stop))
                melodies[part] += chord.score[part] & offset
            else:
                start = idx_stops[part][-1][-1] if len(idx_stops[part]) > 0 else 0
                stop = start + chord.duration
                idx_stops[part].append((start, stop))
                melodies[part] += Silence(chord.duration)

    chord = first_chord(**melodies)
    return chord, idx_stops, offsets, melodies, chords


def get_melody_starting_between(melody, start, stop):
    """

    Parameters
    ----------
    melody :
        
    start :
        
    stop :
        

    Returns
    -------

    """

    new_melody = None
    time = 0
    for note in melody.notes:
        if stop > time >= start:
            new_melody += note.copy()
        if time >= stop:
            break
        time += note.duration

    return new_melody

def note_duration_still_playing_at(melody, start):
    """

    Parameters
    ----------
    melody :
        
    start :
        

    Returns
    -------

    """
    result = None
    time = 0

    for note in melody.notes:
        if (time < start) and (time + note.duration > start):
            result = time + note.duration - start
            return result
        if time >= start:
            return result
        time += note.duration

    return result

def note_still_playing_at(melody, start):
    """

    Parameters
    ----------
    melody :
        
    start :
        

    Returns
    -------

    """
    time = 0
    for note in melody.notes:
        if (time < start) and (time + note.duration > start):
            result = time + note.duration - start
            to_return = note.copy()
            to_return.duration = result
            return to_return
        if time >= start:
            return None
        time += note.duration

    return None


def get_melody_between(melody, start, stop, continuation=True):
    """

    Parameters
    ----------
    melody :
        
    start :
        
    stop :
        
    continuation :
         (Default value = True)

    Returns
    -------

    """
    melody_between = get_melody_starting_between(melody, start, stop)
    if continuation:
        note_still_playing = note_duration_still_playing_at(melody, start)
        if note_still_playing is not None:
            note_still_playing = Continuation(note_still_playing)
    else:
        note_still_playing = note_still_playing_at(melody, start)

    new_melody = None

    if note_still_playing is not None:
        new_melody = note_still_playing
    if melody_between is not None:
        new_melody += melody_between.copy()
    duration = stop - start
    delta_duration = new_melody.duration - duration
    assert delta_duration >= 0, "Delta duration should be positive, error of implementation"
    assert (new_melody.notes[-1].duration - delta_duration) > 0, "Error of implementation {} {}".format(start, stop)
    new_melody.notes[-1].duration = new_melody.notes[-1].duration - delta_duration
    return new_melody



def reproject_on_multiple_chords(chords, new_chord, idx_stops, offsets, offset=True):
    """

    Parameters
    ----------
    chords :
        
    new_chord :
        
    idx_stops :
        
    offsets :
        
    offset :
         (Default value = True)

    Returns
    -------

    """
    new_score = None
    # Put everything back on same melody
    #accs = {key: 0 for key in idx_stops.keys()}
    for idx, chord in enumerate(chords):
        # Get each part of chord
        res_chord = {}
        for part in chord.parts:
            start, stop = idx_stops[part][idx]
            melody = get_melody_between(new_chord.get_part(part), start, stop)
            res_chord[part] = melody
            if offset:
                res_chord[part] = res_chord[part] & - offsets[idx]

        new_score += chord(**res_chord)

    return new_score

def parse_relative_to_absolute(melody):
    """

    Parameters
    ----------
    melody :
        

    Returns
    -------

    """
    result = None
    prev = None
    for note in melody.copy().notes:
        if note.type in ['su', 'sd']:
            to_add = prev.add_interval(note)
            to_add.duration = note.duration
            result += to_add
            prev = to_add.copy()
        elif note.type in ['s', 'a']:
            result += note.copy()
            prev = note.copy()
        elif note.type in ['r', 'l']:
            result += note.copy()
        elif note.type in ['h']:
            DICT_NOTES = {0: 0,
                          1: 1,
                          2: 1,
                          3: 2,
                          4: 2,
                          5: 3,
                          6: 3,
                          7: 4,
                          8: 5,
                          9: 5,
                          10: 6,
                          11: 6
                          }
            new_note = note.copy()
            new_note.type = "s"
            new_note.val = DICT_NOTES[new_note.val]
            result += new_note
        else:
            raise Exception('Could not handle type in project rhythm : {}'.format(note.type))

    return result


def is_continuation(note, silence_as_continuation=True):
    """

    Parameters
    ----------
    note :
        
    silence_as_continuation :
         (Default value = True)

    Returns
    -------

    """
    if silence_as_continuation:
        return note.type == "l" or note.type == "r"
    else:
        return note.type == "l"

def is_rest(note):
    """

    Parameters
    ----------
    note :
        

    Returns
    -------

    """
    return note.type == "r"

def is_in_time(start, end, time):
    """

    Parameters
    ----------
    start :
        
    end :
        
    time :
        

    Returns
    -------

    """
    return (start <= time) and (end > time)


def project_on_rhythm(rhythm, melody):
    """

    Parameters
    ----------
    rhythm :
        
    melody :
        

    Returns
    -------

    """
    new_melody = None

    melody_without_relative = parse_relative_to_absolute(melody.copy())
    notes_ends = np.cumsum([note.duration for note in melody_without_relative.notes]).tolist()
    notes_starts = [0] + notes_ends[:-1]
    notes_times = [(note, start, end) for note, start, end in zip(melody_without_relative.notes, notes_starts, notes_ends)]

    time = 0
    previous_note = None

    for rhythm_note in rhythm.notes:
        # Find first note
        candidates = [note for note, start, end in notes_times if is_in_time(start, end, time)]
        if len(candidates) > 0:
            candidate = candidates[0]
            if candidate.type == 'l':
                if previous_note is None:
                    to_add = Continuation(rhythm_note.duration)
                else:
                    to_add = previous_note.copy()
                    to_add.duration = rhythm_note.duration
            elif rhythm_note.type == 'l':
                to_add = Continuation(rhythm_note.duration)
            else:
                to_add = candidate.copy()
                to_add.duration = rhythm_note.duration
            new_melody += to_add
            if to_add.type != 'l':
                previous_note = to_add.copy()
        else:
            new_melody += Silence(rhythm_note.duration)
        time = time + rhythm_note.duration

    return new_melody


def get_nearest_note(candidates, note):
    """

    Parameters
    ----------
    candidates :
        
    note :
        

    Returns
    -------

    """
    candidate_val = min(candidates, key=lambda x: abs(x.val - note.val))
    return get_nearest_val(note, candidate_val.val)

def get_nearest_val(n, new_val):
    """

    Parameters
    ----------
    n :
        
    new_val :
        

    Returns
    -------

    """
    delta_val = (new_val - n.val) % 7
    notes_candidate = [n.add_value(delta_val, 0), n.add_value(delta_val, -1), n.add_value(delta_val, 1)]
    candidate = min(notes_candidate, key=lambda x: note_distance_abs(n, x))
    return candidate


def note_distance_abs(n1, n2):
    """

    Parameters
    ----------
    n1 :
        
    n2 :
        

    Returns
    -------

    """
    return abs(n2.scale_pitch - n1.scale_pitch)


def note_distance(n1, n2):
    """

    Parameters
    ----------
    n1 :
        
    n2 :
        

    Returns
    -------

    """
    return n1.val - n2.val


def get_nearest_note_in_context(rhythm_note, note, chord, dict_notes=None):
    """

    Parameters
    ----------
    rhythm_note :
        
    note :
        
    chord :
        
    dict_notes :
         (Default value = None)

    Returns
    -------

    """
    if dict_notes is None:
        dict_notes = {}
    candidates = get_notes_candidate_in_context(rhythm_note, chord, dict_notes=dict_notes)
    if len(candidates) == 0:
        return note
    elif note.is_note:
        return get_nearest_note(candidates, note)
    else:
        return note


def get_notes_candidate_in_context(rhythm_note, chord, dict_notes=None):
    """

    Parameters
    ----------
    rhythm_note :
        
    chord :
        
    dict_notes :
         (Default value = None)

    Returns
    -------

    """
    if dict_notes is None:
        dict_notes = {}
    rhythm_note = rhythm_note.set_duration(Q)
    if rhythm_note in dict_notes.keys():
        return dict_notes[rhythm_note]
    else:
        if rhythm_note.val == CHORD_NOTE.val:
            return chord.possible_notes
        elif rhythm_note.val == SCALE_NOTE.val:
            return chord.scale_notes
        elif rhythm_note.val == TONIC.val:
            return [Note("s", 0, 0, 1)]
        elif rhythm_note.val == TONIC_OR_FIFTH.val:
            return [Note("s", 0, 0, 1), Note("s", 4, 0, 1)]
        elif rhythm_note.val == SCALE_DISSONNANCE.val:
            return chord.scale_dissonances
    return [rhythm_note]





def get_absolute_voice(voice):
    """Transform a relative melody to an absolute one

    Parameters
    ----------
    voice :
        return:

    Returns
    -------

    """
    return parse_relative_to_absolute(voice)

def get_absolute_voices(voices):
    """

    Parameters
    ----------
    voices :
        

    Returns
    -------

    """
    return [get_absolute_voice(voice) for voice in voices]


"""
Copyright (c) 2023, Florian GARDIN
All rights reserved.

This source code is licensed under the BSD-style license found in the
LICENSE file in the root directory of this source tree.
"""

# Groupby
from fractions import Fraction as frac


def sequence_to_score(sequence, sort_by_time=True, **kwargs):
    """Convert a dataframe (called a sequence) to a MusicLang score

    Parameters
    ----------
    sequence :
        pandas.DataFrame with required columns
    sort_by_time :
        boolean, default=True, If true sort by 'start' column before reconverting to score
    **kwargs :
        

    Returns
    -------
    type
        score, Score : converted score

    """
    from ..chord import Chord
    from ..tonality import Tonality
    from ..note import Note, Silence, Continuation
    if sort_by_time:
        sequence = sequence.sort_values(by='start', ascending=True)
    groups_chord = sequence.groupby('chord_idx')
    score = None
    for chord_idx, group_chord in groups_chord:
        chord_row = group_chord.iloc[0]
        tonality = Tonality(int(chord_row['tonality_degree']), mode=chord_row['tonality_mode'],
                            octave=int(chord_row['tonality_octave']))
        chord = Chord(int(chord_row['chord_degree']), extension=chord_row['chord_extension'], tonality=tonality)
        parts = {}
        groups_instrument = group_chord.groupby('instrument')
        for instrument, row in groups_instrument:
            part = None
            for idx_note, note in row.iterrows():
                if note['silence']:
                    new_note = Silence(note['note_duration'])
                elif note['continuation']:
                    new_note = Continuation(note['note_duration'])
                else:
                    new_note = Note(note['note_type'],
                                val=int(note['note_val']),
                                octave=int(note['note_octave']),
                                duration=frac(note['note_duration']).limit_denominator(8),
                                amp=int(note['note_amp']))
                part += new_note
            parts[instrument] = part
        score += chord(**parts)
    return score


def score_to_sequence(score, **kwargs):
    """Convert a score to a sequence (pandas.DataFrame with specific columns)

    Parameters
    ----------
    score :
        musiclang.Score
    **kwargs :
        

    Returns
    -------
    type
        sequence, pandas.DataFrame : converted sequence

    """

    sequence = []
    time = 0
    for idx, chord in enumerate(score):
        sequence += [[time, idx] + s for s in chord.to_sequence()]
        time += chord.duration

    import pandas as pd
    import numpy as np

    sequence = pd.DataFrame(sequence,
                            columns=['chord_time', 'chord_idx', 'chord_relative_start', 'chord_relative_end', 'pitch',
                                     'chord', 'instrument', 'note'])
    sequence['start'] = sequence['chord_time'] + sequence['chord_relative_start']
    sequence['end'] = sequence['chord_time'] + sequence['chord_relative_end']
    sequence['chord_degree'] = sequence['chord'].apply(lambda x: x.element)
    sequence['chord_extension'] = sequence['chord'].apply(lambda x: x.extension)
    sequence['chord_octave'] = sequence['chord'].apply(lambda x: x.octave)
    sequence['tonality_degree'] = sequence['chord'].apply(lambda x: x.tonality.degree)
    sequence['tonality_mode'] = sequence['chord'].apply(lambda x: x.tonality.mode)
    sequence['tonality_octave'] = sequence['chord'].apply(lambda x: x.tonality.octave)

    sequence['silence'] = sequence['note'].apply(lambda x: x.is_silence)
    sequence['continuation'] = sequence['note'].apply(lambda x: x.is_continuation)
    sequence['note_type'] = sequence['note'].apply(lambda x: x.type)
    sequence['note_val'] = sequence['note'].apply(lambda x: x.val)
    sequence['note_octave'] = sequence['note'].apply(lambda x: x.octave)
    sequence['note_amp'] = sequence['note'].apply(lambda x: x.amp)
    sequence['note_duration'] = sequence['note'].apply(lambda x: x.duration)
    sequence['note_idx'] = np.arange(len(sequence))
    COLUMNS = ['chord_time', 'chord_idx', 'chord_relative_start', 'chord_relative_end', 'pitch', 'instrument',
                'start', 'end', 'chord_degree', 'chord_extension', 'chord_octave', 'tonality_degree',
               'tonality_mode', 'tonality_octave', 'silence', 'continuation',
               'note_type', 'note_val', 'note_octave', 'note_amp',
               'note_duration', 'note_idx'
               ]
    return sequence[COLUMNS].sort_values(by='start', ascending=True)


import glob
import shutil
import os
import random
class SplitTrainTest:


    def __init__(self,
                 input_pattern,
                 output_directory,
                 train_ratio=0.95,
                 padding=False,
                 padding_size=1024,
                 padding_token="[PAD]",
                 end_of_text_token="[END]",
                 start_of_text_token="[START]"):

        self.input_pattern = input_pattern
        self.output_directory = output_directory
        self.train_ratio = train_ratio
        self.padding_size = padding_size
        self.padding_token = padding_token
        self.end_of_text_token = end_of_text_token
        self.start_of_text_token = start_of_text_token
        self.padding = padding

    def split(self):
        try:
            shutil.rmtree(self.output_directory)
        except:
            pass
        os.makedirs(self.output_directory, exist_ok=True)
        self.create_train_val_split(self.input_pattern, self.output_directory, self.train_ratio)
        """
        Create train and val datasets
        """
        # Create train and val dataset
        train_output = os.path.join(self.output_directory, "train")
        val_output = os.path.join(self.output_directory, "val")
        self.prepare_dataset(train_output, os.path.join(self.output_directory, 'train.txt'))
        self.prepare_dataset(val_output, os.path.join(self.output_directory, 'val.txt'))

    def prepare_dataset(self, directory, output_file):
        file_pattern = os.path.join(directory, "*.txt")
        files = glob.glob(file_pattern)
        end_of_text = self.end_of_text_token
        texts = []
        for file in files:
            with open(file, encoding="utf-8") as f:
                text = f.read()
                text = text.replace('\n', '')
            # Add start-of-text and end-of-text symbols
            if self.padding:
                padding = "".join([self.padding_token] * (self.padding_size - 1))
                text = padding + self.start_of_text_token + text + end_of_text
            else:
                text = self.start_of_text_token + text + end_of_text

            texts.append(text)

        full_text = "\n".join(texts)
        full_text = full_text.replace('\t', '')
        full_text = full_text.replace(' ', '')
        with open(output_file, 'w') as f:
            f.write(full_text)

    def create_train_val_split(self, input_directory, output_directory, train_val_ratio=0.8):
        file_pattern = input_directory
        files = glob.glob(file_pattern)
        random.shuffle(files)

        train_size = int(len(files) * train_val_ratio)
        train_files = files[:train_size]
        val_files = files[train_size:]

        train_output = os.path.join(output_directory, "train")
        val_output = os.path.join(output_directory, "val")

        os.makedirs(train_output, exist_ok=True)
        os.makedirs(val_output, exist_ok=True)

        for file in train_files:
            shutil.copy(file, os.path.join(train_output, os.path.basename(file)))

        for file in val_files:
            shutil.copy(file, os.path.join(val_output, os.path.basename(file)))



from musiclang.transform import TransformGraph


from musiclang.transform.features import ExtractMainTonality
from musiclang.transform.score_merger import ConcatWithPattern, ConcatScores
from musiclang.transform.chord import Modulate


class SonataGraph(TransformGraph):
    """
    Create a simple sonata from two themes (A, B) in their own tonalities
    - Returns the pattern : A, B, A, B in A tonality
    """
    NODES = [
        ('A_TONALITY', ['A'], ExtractMainTonality()),
        ('B_TONALITY', ['B'], ExtractMainTonality()),
        ('B_TO_A', ['A_TONALITY', 'B_TONALITY'], lambda x, y: x - y),
        ('B_IN_A_TONALITY', ['B', 'B_TO_A'], lambda score, a_tone: Modulate(a_tone)(score)),
        ('Sonata', ['A', 'B', 'B_IN_A_TONALITY'], ConcatWithPattern([0, 1, 0, 2]))
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(['A', 'B'], 'Sonata', nodes=self.NODES)


class RondoGraph(TransformGraph):

    """
    Create a Rondo form from a A-theme and other themes (how many you want). Finish by the A-theme
    """

    def __init__(self, a_theme, *other_themes, **kwargs):
        inputs = [a_theme] + list(other_themes)
        themes = [a_theme]
        for theme in other_themes:
            themes.append(theme)
            themes.append(a_theme)

        nodes = [('Rondo', themes, ConcatScores())]
        super().__init__(inputs, 'Rondo', nodes=nodes)


from .write.score import Score
from .write.note import Note, Silence, Continuation
from .write.melody import Melody
from .write.chord import Chord
from .write.tonality import Tonality
from .write.element import Element
from .write.rhythm import Metric, CompositeMetric, ScoreRhythm

from .write import library
from .analyze import ScoreFormatter
from .transform.library import VoiceLeading
from .transform import PartComposer

__all__ = ['Score', 'Note', 'Silence', 'Continuation', 'Melody',
           'Chord', 'Tonality', 'Element', 'library', 'Metric', 'CompositeMetric', 'ScoreRhythm', 'ScoreFormatter',
           'VoiceLeading', 'PartComposer'
           ]

__author__ = 'Florian GARDIN'



from musiclang.transform.composing import project_on_score_keep_notes
from musiclang.library import *

def test_project_on_score_keep_notes():


    score1 = (I % I.M)(s0 + s0)
    score2 = (I % I.M)(s4) + (IV % I.M)(s4)  # Whatever note, only the harmonic rythm counts here
    expected_score = (I % I.M)(s0) + (IV % I.M)(s4.o(-1))
    new_score = project_on_score_keep_notes(score1, score2)

    assert new_score == expected_score


def test_project_on_score_keep_notes_with_continuation():


    score1 = (I % I.M)(s0 + s0)
    score2 = (I % I.M)(s4.e) + (IV % I.M)(s4.qd)  # Whatever note, only the harmonic rythm counts here
    expected_score = (I % I.M)(s0.e) + (IV % I.M)(l.e + s4.o(-1))
    new_score = project_on_score_keep_notes(score1, score2)

    assert new_score == expected_score


def test_project_on_score_keep_notes_different_durations_keep_min_duration():

    score1 = (I % I.M)(s0 + s0)
    score2 = (I % I.M)(s4.e) + (IV % I.M)(s4.h)  # Whatever note, only the harmonic rythm counts here
    expected_score = (I % I.M)(s0.e) + (IV % I.M)(l.e + s4.o(-1))
    new_score = project_on_score_keep_notes(score1, score2)

    assert new_score == expected_score




def best_voice_leading(fpd, pcd):
    pass


import numpy as np

class Rules:

    PARALLEL_OCTAVES = 'parallel_octaves'
    PARALLEL_FIFTHS = 'parallel_fifths'
    PARALLEL_DISSONNANCES = 'parallel_dissonnances'
    CROSSING = 'no_crossing'
    UNISSON = 'no_unisson'


ALL_RULES = [Rules.PARALLEL_OCTAVES, Rules.PARALLEL_FIFTHS,
             Rules.PARALLEL_DISSONNANCES, Rules.CROSSING,
             Rules.UNISSON]


class VoiceLeading:
    """
    This class is responsible of optimizing a voice leading on the first notes of a chord progression.

    - It first tries to minimize the movement of voices between each chords
    - It then finetune the solution using classic voice movement rules that are parametrizable (parallel octaves, fifths...)

    .. warning:: We strongly advise to first derive a good chord progression, then apply voice leading algorithm, then apply rythms
    and melodies because this class only deals with the first note of a chord.

    Examples
    --------
    >>> from musiclang.library import *
    >>> from musiclang.transform import VoiceLeading

    >>> score = (I % I.M)(cello=s0, violin=s1) * 2
    >>> vl = VoiceLeading(fixed_voices=['cello__0'], seed=5)
    >>> new_score = vl.optimize(score, method='voices_and_rules', max_iter=1000, max_iter_rules=100, max_norm=3)
    >>> print(new_score)
    (I % I.M)(cello=s0, violin=s1) + (I % I.M)(cello=s0, violin=s2)

    It has automatically solved the parallel dissonnance issue.
    """

    TYPES = ['b', 'c', 's', 'h']
    RULES = Rules

    def __init__(self, types=None, rules=None, exclude_rules=None, fixed_voices=None, change_octave_fixed=True, seed=34, method='voices_and_rules', **kwargs):
        """

        Parameters
        ----------
        types: None or list[str], default will be ['b', 'c', 's', 'h']
            List of types on which to apply the algorithm, if not present the notes will act as fixed voices
        rules: list of str in ALL_RULES
            Rules to apply if you use "voices_and_rules" or "rules" optimizer
            Existing rules are :
            - parallel_octaves
            - parallel_fifths
            - parallel_dissonnances
        fixed_voices: list[str]
            List of name of voice (with voice idx like piano__0) to not modify
        change_octave_fixed: list[str] or boolean or None
            Do we apply the octave change for the fixed voices (useful to set to true for bass)
        seed: int
            Seed used for random generation to get reproducible results
        """
        self.types = types if types is not None else self.TYPES
        self.fixed_voices = [] if (fixed_voices is None) else fixed_voices
        self.change_octave_fixed = change_octave_fixed
        if isinstance(self.change_octave_fixed, bool):
            self.change_octave_fixed = [self.change_octave_fixed for i in self.fixed_voices]

        self.rules = rules if rules is not None else ALL_RULES
        self.seed = seed
        self.rg = np.random.RandomState(seed)
        self.method = method
        self.kwargs = kwargs
        if exclude_rules is not None:
            self.rules = [r for r in self.rules if r not in exclude_rules]

    def optimize(self, score, **kwargs):
        """
        Main entry point to optimize a score

        Parameters
        ----------
        score: Score
            Musiclang score on which to apply voice leading algorithm
        method: str
            Method, existing are : "voices", "voices_and_rules", "rules", "random"
            - voices : Try to create a parcimonious voice leading, starting with the proposed solution
            - rules : Try to erase voices movements errors (like parallel octaves or fifths)
            - voices_and_rules: Combo of both
            - random: Same metric as "voices" but the optimisation algorithm is completely random, use only if you know what you do

        kwargs: **kwargs
            Optimizer parameters
            - max_iter: Number of iteration of parcimonious voices optimizer
            - max_iter_rules: Number of iteration of harmonic rules enforcer
            - max_norm: Max movement allowed in term of relative value of note, set low if you want minimal changes
            - temperature: Set high for high degree of randomness
            - min_temperature: Temperature at the end of the voice optimisation (only for voice optimisation)

        Returns
        -------

        """
        self.init(score)
        self.rg = np.random.RandomState(self.seed)
        dvals = np.zeros(self.pitch.shape, dtype=np.int16)
        # DO OPTIMIZATION HERE
        #solution = self.random_optim(dvals, **kwargs)
        if self.method == 'voices_and_rules':
            solution = self.voices_and_rules(dvals, **self.kwargs)
        elif self.method == 'voices':
            solution = self.voices_optim(dvals, **self.kwargs)
        elif self.method == 'rules':
            solution = self.optimize_rules(dvals, **self.kwargs)
        elif self.method == 'random':
            solution = self.random_optim(dvals, **self.kwargs)
        else:
            raise Exception('Not existing method, existing are : "voices", "voices_and_rules", "rules", "random"')
        #######################

        return self.get_score(score, solution)

    def find_optimal_octaves(self, score):
        """
        Find optimal chord octaves, to get the most constant bass pitch
        Parameters
        ----------
        score

        Returns
        -------

        """
        def recursive_correct_octave(chord):
            bass_pitch = chord.bass_pitch
            if bass_pitch > 6:
                new_chord = chord.o(-1)
                for voice, change in zip(self.fixed_voices, self.change_octave_fixed):
                    if not change:
                        new_chord.score[voice] = new_chord.score[voice].o(1)
                return recursive_correct_octave(new_chord)
            elif bass_pitch <= -6:
                new_chord = chord.o(1)
                for voice, change in zip(self.fixed_voices, self.change_octave_fixed):
                    if not change:
                        new_chord.score[voice] = new_chord.score[voice].o(-1)
                return recursive_correct_octave(new_chord)
            else:
                return chord

        new_score = None
        for chord in score.chords:
            new_score += recursive_correct_octave(chord)
        return new_score

    def __call__(self, score, **kwargs):
        # First Find best chords octaves
        score = self.find_optimal_octaves(score)
        new_score = self.optimize(score, **kwargs)
        return new_score

    def init(self, score):
        """
        Given existing chord progression find the best voice leading on given types, don't change other
        kinds of notes.
        Apply only for the first note of melodies chords

        Parameters
        ----------
        score: Score
            Score on which to apply the voice leading

        types: list[str]
            Types on which to apply a voice leading

        """
        import numpy as np

        score = score.to_score(copy=False).normalize_instruments()
        self.instruments = score.instruments

        pitches_chords = [chord.chord_pitches for chord in score.chords]
        pitches_extensions = [chord.chord_extension_pitches for chord in score.chords]
        pitches_scale = [chord.scale_pitches for chord in score.chords]
        pitches_chromatic = [chord.chromatic_scale_pitches for chord in score.chords]
        pitches_absolute = list(range(12))
        self.candidates_raw = [{
            'b': pitches_extensions[i],
            'c': pitches_chords[i],
            's': pitches_scale[i],
            'h': pitches_chromatic[i],
            'a': pitches_absolute,
            'r': [0],
            'l': [0],
            'bass': pitches_extensions[i][0]
        } for i in range(len(score.chords))]

        self.array = [[chord.score[ins].notes[0] for chord in score.chords]
                      for ins in self.instruments]
        self.type = np.asarray([[i.type for i in j] for j in self.array])
        self.val = np.asarray([[i.val for i in j] for j in self.array]).astype(np.int16)
        self.octave = np.asarray([[i.octave for i in j] for j in self.array]).astype(np.int16)
        self.pitch = np.asarray([[chord.to_pitch(chord.score[ins].notes[0]) or 0 for chord in score.chords]
                                 for ins in self.instruments]).astype(np.int16)

        idxs_instruments = [self.instruments.index(f) for f in self.fixed_voices]
        self.dvalsmask = np.ones(self.pitch.shape, dtype=np.int16)
        # Mask instruments
        for idx in idxs_instruments:
            self.dvalsmask[idx, :] = 0
        # Mask silences and continuations
        type_mask = np.zeros(self.dvalsmask.shape)
        for type in self.types:
            type_mask += 1 * (self.type == type)
        type_mask = 1 * (type_mask > 0)
        self.dvalsmask *= type_mask.astype(np.int16)
        self.candidates = [[self.candidates_raw[idxc][self.type[idxi, idxc]]
                            for idxc, chord in enumerate(score.chords)]
                           for idxi, instrument in enumerate(self.instruments)]

        return score

    def get_candidate_at(self, idx_ins, idx_chord):
        return self.candidates[idx_ins][idx_chord]

    def get_current_pitch_at(self, idx_ins, idx_chord):
        return self.pitch[idx_ins, idx_chord]

    def get_pitch_solution(self, dvals):
        new_vals = self.val + dvals
        octs = self.octave
        pitches = np.asarray([[self.candidates[idxi][idxc][new_val % len(self.candidates[idxi][idxc])]
                               + 12 * (new_val // len(self.candidates[idxi][idxc]))
                               for idxc, new_val in enumerate(row)]
                              for idxi, row in enumerate(new_vals)])
        pitches += 12 * octs

        return pitches

    def eval_solution(self, dvals):
        pitches = self.get_pitch_solution(dvals)
        # Minimize movement
        movement = self.get_movement(pitches)
        movement_score = np.mean(np.power(np.abs(movement), 1))
        # space = np.diff(pitches, axis=0)
        # space_score =  np.mean(np.power(space, 2))
        return movement_score

    def get_movement(self, pitches):
        movement = np.diff(pitches, axis=1)
        return movement



    def random_optim(self, dvals, max_iter=250, max_norm=3, **kwargs):
        solutions = [dvals + self.rg.randint(-max_norm, max_norm + 1, dvals.shape) for i in range(max_iter - 1)]
        solutions.append(dvals)
        scores = [self.eval_solution(sol) for sol in solutions]
        best = np.argmin(scores)
        return solutions[best]

    def voices_optim(self, dvals, max_iter=100, max_norm=3, temperature=8, min_temperature=1, **kwargs):
        min_score = self.eval_solution(dvals)
        start_temp = temperature
        end_temp = min_temperature
        for it in range(max_iter):
            pitches = self.get_pitch_solution(dvals)
            mov = self.get_movement(pitches).astype(np.float32)  # Find movements
            mov *= self.dvalsmask[:, :-1]
            sgn_mov = np.sign(mov)
            # Find probabilities of moving an index
            proba = np.exp(mov/temperature)/np.sum(np.exp(mov/temperature))
            assert np.max(proba) <= 1
            # Sample index given proba
            delta1 = sgn_mov * (self.rg.random(mov.shape) < proba)
            delta1 = (np.c_[delta1, np.zeros(len(delta1))])
            delta2 = - sgn_mov * (self.rg.random(mov.shape) < proba)
            delta2 = (np.c_[np.zeros(len(delta2)), delta2])
            mov = (delta1 + delta2).astype(int)
            proposed_sol = (dvals + mov).clip(-max_norm, max_norm)
            score = self.eval_solution(proposed_sol)
            if score < min_score:
                min_score = score
                dvals = proposed_sol
            coeff = (it / max_iter)
            temperature = coeff * end_temp + (1 - coeff) * start_temp
        return dvals
        # Move in direction of sgn


    def optimize_rules(self, dvals, max_iter_rules=100, max_norm_rules=1, temperature=1, **kwargs):

        def eval_solution(dvals):
            pitches = self.get_pitch_solution(dvals)
            problems = self.get_problems(pitches)
            return np.sum(problems)

        pitches = self.get_pitch_solution(dvals)
        problems = self.get_problems(pitches)
        min_score = eval_solution(dvals)
        tried = set()
        print(temperature)
        for i in range(max_iter_rules):
            pitches = self.get_pitch_solution(dvals)
            problems = self.get_problems(pitches)
            noise = 0.0
            proba = np.exp((problems + noise)/temperature)/np.sum(np.exp((problems + noise)/temperature))
            m, M = -1, 2
            delta = self.rg.randint(-max_norm_rules, max_norm_rules+1, proba.shape) * (self.rg.random(proba.shape) < proba)
            delta = delta.clip(-max_norm_rules, max_norm_rules)
            proposed_sol = (dvals + delta)
            proposed_sol *= self.dvalsmask
            proposed_set = tuple(proposed_sol.flatten().tolist())
            if proposed_set in tried:
                continue
            else:
                tried.add(proposed_set)

            score = eval_solution(proposed_sol)
            if score < min_score:
                min_score = score
                dvals = proposed_sol
            if min_score == 0:
                break
        score = self.eval_solution(dvals)
        print('Problem score :', min_score)
        print('SCORE after tuning :', score)
        return dvals

    def get_val(self, pitches, v1, v2, idxc):
        if idxc >= pitches.shape[1]:
            return -1
        if self.type[v1, idxc] in ['r', 'l']:
            return -1
        if self.type[v2, idxc] in ['r', 'l']:
            return -1

        return abs(pitches[v1, idxc] - pitches[v2, idxc]) % 12

    def get_problems(self, pitches):
        # Parallel only counts if movement is not null
        movement = self.get_movement(pitches)
        moving = movement != 0
        dissonnances = {1, 2, 6, 11, 10}
        # Find intervals between all instruments
        intervals = np.zeros((pitches.shape[0], *pitches.shape))
        abs_intervals = np.zeros((pitches.shape[0], *pitches.shape))
        parallel_fifths = np.zeros(pitches.shape)
        parallel_octaves = np.zeros(pitches.shape)
        parallel_dissonnances = np.zeros(pitches.shape)
        unissons = np.zeros(pitches.shape)
        crossings = np.zeros(pitches.shape)
        for v1 in range(pitches.shape[0]):
            for v2 in range(pitches.shape[0]):
                if v2 >= v1:
                    break
                for idxc in range(pitches.shape[1]):
                    rel_val = pitches[v1, idxc] - pitches[v2, idxc]
                    abs_val = abs(rel_val)
                    val = abs_val % 12
                    intervals[v1, v2, idxc] = val
                    intervals[v2, v1, idxc] = val
                    abs_intervals[v1, v2, idxc] = abs_val
                    if Rules.PARALLEL_FIFTHS in self.rules:
                        if val == 7 and val == self.get_val(pitches, v1, v2, idxc + 1):
                            parallel_fifths[v1, idxc] += 1
                            parallel_fifths[v2, idxc] += 1
                            parallel_fifths[v1, idxc + 1] += 1
                            parallel_fifths[v2, idxc + 1] += 1

                    if Rules.PARALLEL_OCTAVES in self.rules:
                        if val == 0 and val == self.get_val(pitches, v1, v2, idxc + 1):
                            parallel_octaves[v1, idxc] += 1
                            parallel_octaves[v2, idxc] += 1
                            parallel_octaves[v1, idxc + 1] += 1
                            parallel_octaves[v2, idxc + 1] += 1
                    if Rules.PARALLEL_DISSONNANCES in self.rules:

                        if val in dissonnances and self.get_val(pitches, v1, v2, idxc + 1) in dissonnances:
                            parallel_dissonnances[v1, idxc] += 1
                            parallel_dissonnances[v2, idxc] += 1
                            parallel_dissonnances[v1, idxc + 1] += 1
                            parallel_dissonnances[v2, idxc + 1] += 1
                    if Rules.CROSSING in self.rules:
                        if rel_val < 0:
                            crossings[v1, idxc] += 3
                            crossings[v2, idxc] += 3
                    if Rules.UNISSON in self.rules:
                        if abs_val == 0:
                            unissons[v1, idxc] += 5
                            unissons[v2, idxc] += 5

        problems = parallel_fifths + parallel_octaves + parallel_dissonnances + unissons + crossings
        problems *= self.dvalsmask
        return problems


    def voices_and_rules(self, dvals, **kwargs):
        dvals = self.voices_optim(dvals, **kwargs)
        # REMOVE CONSECUTIVE FIFTHS AND OCTAVES
        dvals = self.optimize_rules(dvals, **kwargs)

        return dvals




    def get_corrected_note(self, note, new_vals, idx_ins, idx_chord):
        nb = len(self.get_candidate_at(idx_ins, idx_chord))
        new_val = new_vals[idx_ins, idx_chord]
        note.val = int(new_val % nb)
        note.octave += int(new_val // nb)
        return note

    def get_score(self, score, dvals):
        new_score = score.copy()
        new_vals = self.val + dvals
        for idx_chord, chord in enumerate(new_score.chords):
            for idx_ins, ins in enumerate(self.instruments):
                if ins in chord.score.keys():
                    first_note = chord.score[ins].notes[0]
                    if not first_note.is_silence and not first_note.is_continuation:
                        chord.score[ins].notes[0] = self.get_corrected_note(first_note, new_vals, idx_ins, idx_chord)

        return new_score




import time

from musiclang.write.pitches.pitches_utils import get_relative_scale_value
from musiclang import Note, Melody
import numpy as np
from musiclang.library import *
from musiclang import Score
from musiclang.write.out import get_pitches, get_pitches_instruments


def get_candidate(candidates, scale, new_pitch, last_pitch):
    for candidate in candidates:
        candidate_pitch = get_relative_scale_value(candidate, last_pitch, scale)
        candidate_pitch_class = candidate_pitch % 12
        if candidate_pitch_class == new_pitch % 12:
            delta_octave = (new_pitch - candidate_pitch) // 12

            return candidate.oabs(delta_octave)
    return None


def transform_type_to_relative_note(new_type, scale, new_pitch, last_pitch):
    candidates_up = [Note(new_type + 'u', i, 0, 1) for i in range(len(scale))]
    candidates_down = [Note(new_type + 'd', i, 0, 1) for i in range(len(scale))]
    candidates = candidates_up + candidates_down
    candidate = get_candidate(candidates, scale, new_pitch, last_pitch)
    return candidate


def transform_note_to_relative_note(note, chord, last_pitch):
    new_pitch = chord.to_pitch(note)
    scale = chord.get_scale_from_type(note.type)
    result_note = transform_type_to_relative_note(note.type, scale, new_pitch, last_pitch)
    new_note = note.copy()
    new_note.type = result_note.type
    new_note.val = result_note.val
    new_note.octave = result_note.octave

    return new_note


def transform_note_to_chord_extension(note, chord):
    pitch = chord.to_pitch(note)
    pitch_class = pitch % 12
    chord_pitches = chord.chord_extension_pitches
    chord_pitch_classes = [c % 12 for c in chord_pitches]
    extension_val = chord_pitch_classes.index(pitch_class)
    extension_octave = (pitch - chord_pitches[extension_val]) // 12

    real_note = note.copy()
    real_note.type = 'b'
    real_note.val = extension_val
    real_note.octave = extension_octave
    return real_note


def transform_note_to_relative_note_with_chord_note(note, chord, last_pitch, voicing, idx_instrument):
    if not note.is_note:
        return note, last_pitch, voicing
    # First transform to chord extension note if possible
    chord_pitch_classes = chord.chord_extension_pitch_classes
    new_pitch = chord.to_pitch(note)
    pitch_class = new_pitch % 12
    start = time.time()
    if pitch_class in chord_pitch_classes:
        # Transform to chord extension note
        real_note = transform_note_to_chord_extension(note, chord)

        # If voicing is None, then attribute voicing
        if voicing is None:
            new_note = real_note.copy()
            voicing = new_note.set_duration(1).set_amp('mf')
            new_note.type = "x"
            new_note.val = idx_instrument
            new_note.octave = 0
        else:
            new_note = transform_note_to_relative_note(real_note, chord, last_pitch)

    else:  # It is not a chord extension note
        # Then transform it to relative note
        if last_pitch is None:
            last_pitch = chord.to_pitch(c0)
            base_melody = c0.n
        else:
            base_melody = None
        new_note = transform_note_to_relative_note(note, chord, last_pitch)
        new_note = base_melody + new_note
    return new_note, new_pitch, voicing


def transform_melody_to_relative(melody, chord, idx_instrument, prev_voicing=None):
    last_pitch = None
    new_melody = None
    voicing = None
    new_melody = []
    for note in melody.notes:
        new_note, last_pitch, voicing = transform_note_to_relative_note_with_chord_note(note, chord,
                                                                                        last_pitch, voicing,
                                                                                        idx_instrument)
        if isinstance(new_note, Note):
            new_melody.append(new_note)
        else:
            new_melody += new_note.notes

    new_melody = Melody(new_melody)
    if voicing is None:
        voicing = prev_voicing

    return new_melody, voicing


def exclude_instruments(chord, nb, ascending=True):
    instruments = chord.instruments
    pitches = [np.mean([chord.to_pitch(n) for n in chord.score[ins].notes if n.is_note]) for ins in instruments]
    if ascending:
        instruments = sorted(zip(instruments, pitches), key=lambda x: x[1])
    else:
        instruments = sorted(zip(instruments, pitches), key=lambda x: -x[1])
    if nb > 0:
        instruments = instruments[:-nb]
    parts = [i[0] for i in instruments]
    new_chord = chord(**{instr: item for instr, item in chord.score.items() if instr in parts})

    return new_chord, parts


def inverse_recursive_correct_octave(chord):
    """
    Transform a chord to a chord with bass pitch between -6 and 6
    Parameters
    ----------
    chord

    Returns
    -------

    """
    bass_pitch = chord.bass_pitch
    if bass_pitch > 6:
        chord = chord.o(-1)
        for voice, melody in chord.score.items():
            chord.score[voice] = melody.o(1)
        return inverse_recursive_correct_octave(chord)
    elif bass_pitch <= -6:
        chord = chord.o(1)
        for voice, melody in chord.score.items():
            chord.score[voice] = melody.o(-1)
        return inverse_recursive_correct_octave(chord)
    else:
        new_chord = chord.copy()
        return new_chord

def recursive_correct_octave(chord):
    """
    Transform chord and melodies to normalize chord octaves
    Parameters
    ----------
    chord

    Returns
    -------

    """
    bass_pitch = chord.bass_pitch
    if bass_pitch > 6:
        chord = chord.o(-1)
        for voice, melody in chord.score.items():
            chord.score[voice] = melody.o(1)
        return recursive_correct_octave(chord)
    elif bass_pitch <= -6:
        chord = chord.o(1)
        for voice, melody in chord.score.items():
            chord.score[voice] = melody.o(-1)
        return recursive_correct_octave(chord)
    else:
        new_chord = chord.copy()
        new_chord.octave = 0
        return new_chord

def transform_chord_to_relative(chord, nb_instruments_excluded=0, ascending=True):
    chord = recursive_correct_octave(chord)
    chord, parts = exclude_instruments(chord, nb_instruments_excluded, ascending=ascending)
    voicings = []
    prev_voicing = b0.o(-2)
    score_dict = {}
    instruments = [instr.split('__')[0] for instr in parts]
    nb_instruments = len(chord.score.items())
    for idx, part in enumerate(parts):
        melody = chord.score[part]
        new_melody, voicing = transform_melody_to_relative(melody, chord, idx, prev_voicing)
        voicings.append(voicing)
        prev_voicing = voicing
        score_dict[f'v__{idx}'] = new_melody

    result_score = NC(**score_dict)
    bar_duration = result_score.duration

    return result_score, voicings, instruments, bar_duration


class PatternExtractor:
    """
    Extract a pattern that can be auto reharmonized from a chord
    """

    def __init__(self, nb_excluded_instruments=0,
                 fixed_bass=True, voice_leading=True,
                 melody=False,
                 instruments=None,
                 voicing=None
                 ):
        """

        Parameters
        ----------
        nb_excluded_instruments: int
            Number of instruments to exclude from the pattern (starting from highest pitched instrument)
        fixed_bass: bool
            If True, the bass is fixed but the other instruments can be reharmonized
        voice_leading: bool
            If True, the voice leading will be optimized
        melody: bool
            If True, the pattern will extract only the melody
        instruments: list
            List of instruments to use for the pattern
        voicing
        """
        self.nb_excluded_instruments = nb_excluded_instruments
        self.fixed_bass = fixed_bass
        self.voice_leading = voice_leading
        self.instruments = instruments
        self.voicing = voicing
        self.melody = melody

    def extract(self, chord):

        if self.melody:
            result_score, voicing, instruments, bar_duration = transform_chord_to_relative(chord,
                                                                                            len(chord.score.items()) - 1,
                                                                                                ascending=False)
        else:
            start = time.time()
            result_score, voicing, instruments, bar_duration = transform_chord_to_relative(chord,
                                                                                            self.nb_excluded_instruments,
                                                                                            ascending=True)
            print('transform chord to relative', time.time() - start)

        if self.instruments is not None:
            instruments = self.instruments[:len(instruments)] + instruments[len(self.instruments):]
        if self.voicing is not None:
            voicing = self.voicing[:len(voicing)] + voicing[len(self.voicing):]

        dict_pattern = {
            "orchestra": {
                "bar_duration": bar_duration,
                "nb_instruments": len(instruments),
                "pattern": result_score
            },
            "voicing": [note for note in voicing],
            "fixed_bass": self.fixed_bass and not self.melody,
            "voice_leading": self.voice_leading,
            "instruments": instruments
        }

        return dict_pattern

class PatternFeatureExtractor:

    """
    Extract features from a pattern dictionary.
    Returns a dictionary with the features
    """
    def __init__(self):
        pass

    def add_metadata(self, dict_pattern, chord_idx, chord, nb_excluded_instruments, melody):
        metadata = {
            "chord_idx": chord_idx,
            "chord": chord.to_chord(),
            "melody": melody,
            "degree": chord.degree,
            "tonality_degree": chord.tonality.degree,
            "extension": str(chord.extension),
            "tonality_mode": chord.tonality.mode,
            "nb_instruments": dict_pattern["orchestra"]["nb_instruments"],
            "nb_excluded_instruments": nb_excluded_instruments,
            "bar_duration": dict_pattern["orchestra"]["bar_duration"],
            "instruments": dict_pattern["instruments"]
        }

        ## ADD PROPERTIES TO DATA
        dict_pattern['metadata'] = metadata
        return dict_pattern


    def get_durations_of_notes(self, pattern):
        return [float(note.duration) for key, instr in pattern.items() for note in instr.notes if note.type not in ['l', 'r']]

    def min_pitch(self, data, score, pitches):
        min_pitch = min(pitches)
        return min_pitch

    def max_pitch(self, data, score, pitches):
        max_pitch = max(pitches)
        return max_pitch

    def average_pitch(self, data, score, pitches):
        if len(pitches) == 0:
            return 0
        average_pitch = np.mean(pitches)
        return average_pitch

    def max_leap(self, data, score, pitches_instrs):
        leaps = [np.max(np.abs(np.diff(np.asarray(pitches)))) for pitches in pitches_instrs if len(pitches) > 1]
        if len(leaps) == 0:
            return 0
        max_leap = max(leaps)

        return max_leap

    def std_leap(self, data, score, pitches_instrs):
        leaps = [np.std(np.asarray(pitches)) for pitches in pitches_instrs if len(pitches) > 1]
        if len(leaps) == 0:
            return 0
        average_leap = np.mean(leaps)
        return average_leap

    def std_duration(self, data, score):
        pattern = data['orchestra']['pattern']
        durations = self.get_durations_of_notes(pattern)
        if len(durations) < 2:
            return 0
        std_duration = np.std(durations)
        return std_duration

    def average_duration(self, data, score):
        pattern = data['orchestra']['pattern']
        durations = self.get_durations_of_notes(pattern)
        if len(durations) < 1:
            return 0
        average_duration = np.mean(durations)
        return average_duration

    def notes_per_quarter(self, data, score):
        pattern = data['orchestra']['pattern']
        area = pattern.duration
        notes = [note for key, instr in pattern.items() for note in instr.notes if note.type not in ['l', 'r']]
        nb_notes = len(notes)

        return nb_notes / area

    def notes_per_quarter_per_instrument(self, data, score):
        pattern = data['orchestra']['pattern']
        area = pattern.duration * len(pattern.items())
        notes = [note for key, instr in pattern.items() for note in instr.notes if note.type not in ['l', 'r']]
        nb_notes = len(notes)

        return nb_notes / area

    def note_density(self, data, score):
        pattern = data['orchestra']['pattern']
        area = pattern.duration * len(pattern.items())
        if area == 0:
            return 0
        notes = [note for key, instr in pattern.items() for note in instr.notes]
        note_area = sum([note.duration for note in notes if note.type not in ['l', 'r']])
        return note_area / area

    def chromatic_density(self, data, score):
        pattern = data['orchestra']['pattern']
        area = pattern.duration * len(pattern.items()) * data['metadata']['note_density']
        if area == 0:
            return 0
        notes = [note for key, instr in pattern.items() for note in instr.notes]
        chromatic_area = sum([note.duration for note in notes if note.type[0] == 'h'])
        return chromatic_area / area

    def scale_density(self, data, score):
        pattern = data['orchestra']['pattern']
        area = pattern.duration * len(pattern.items()) * data['metadata']['note_density']
        if area == 0:
            return 0
        notes = [note for key, instr in pattern.items() for note in instr.notes]
        scale_area = sum([note.duration for note in notes if note.type[0] == 's'])
        return scale_area / area

    def dissonance_density(self, data, score):
        return data['metadata']['scale_density'] + data['metadata']['chromatic_density']

    def consonance_density(self, data, score):
        return 1 - data['metadata']['dissonance_density']

    def chord_density(self, data, score, all_notes):
        pattern = data['orchestra']['pattern']
        area = pattern.duration * len(pattern.items()) * data['metadata']['note_density']
        if area == 0:
            return 0
        notes = all_notes
        chord_area = sum([note.duration for note in notes if note.type[0] in ['x', 'b']])
        return chord_area / area

    def get_features(self, data):
        pattern = data['orchestra']['pattern']
        all_notes = [note for key, instr in pattern.items() for note in instr.notes]
        score = data['orchestra']['pattern'].project_pattern(data['voicing'])
        pitches = get_pitches(score)
        pitches_instrs = get_pitches_instruments(score)

        data['metadata']['min_pitch'] = float(self.min_pitch(data, score, pitches))
        data['metadata']['max_pitch'] = float(self.max_pitch(data, score, pitches))
        data['metadata']['note_density'] = float(self.note_density(data, score))
        data['metadata']['std_duration'] = float(self.std_duration(data, score))
        data['metadata']['notes_per_quarter'] = float(self.notes_per_quarter(data, score))
        data['metadata']['notes_per_quarter_per_instrument'] = float(self.notes_per_quarter_per_instrument(data, score))
        data['metadata']['average_duration'] = float(self.average_duration(data, score))
        data['metadata']['std_leap'] = float(self.std_leap(data, score, pitches_instrs))
        data['metadata']['max_leap'] = float(self.max_leap(data, score, pitches_instrs))
        data['metadata']['average_pitch'] = float(self.average_pitch(data, score, pitches))
        data['metadata']['chromatic_density'] = float(self.chromatic_density(data, score))
        data['metadata']['scale_density'] = float(self.scale_density(data, score))
        data['metadata']['chord_density'] = float(self.chord_density(data, score, all_notes))
        data['metadata']['dissonance_density'] = float(self.dissonance_density(data, score))
        data['metadata']['consonance_density'] = float(self.consonance_density(data, score))

    def extract(self, data, chord, chord_idx=0, nb_excluded_instruments=0, melody=False):
        pattern = self.add_metadata(data,
                                    chord_idx=chord_idx,
                                    chord=chord,
                                    nb_excluded_instruments=nb_excluded_instruments,
                                    melody=melody)

        # Extract features from the score
        self.get_features(pattern)
        features = {}

        # Extract

        return pattern
